/**
 * @file     dsado.js
 * @description
 *           ADOでCSVを操作しJSONで読み出す.
 *           CSVはRFC4180の拡張版を扱う.
 * @auther   Masashi Nakanishi
 * @license  MIT License  http://www.opensource.org/licenses/mit-license.php
 */
;(function($){
	// see:
	// Microsoft Jet SQL リファレンス
	//   http://office.microsoft.com/ja-jp/access-help/CH006252688.aspx
	// ADO API リファレンス
	//   http://msdn.microsoft.com/ja-jp/library/cc408215.aspx

	var messages = {
		'open.error.filenotexist': 'File not exists',
		'open.error.connection': 'Failed to open a connection',
		'setquery.error.brackets': 'unpaired brackets',
		'setquery.error.parentheses': 'unpaired parentheses',
		'setquery.error.emptystmt': 'empty statement',
		'setquery.error.notselect': 'only "SELECT" statements are acceptable',
		'setquery.error.nofrom': '"FROM" not found',
		'setquery.error.nocolumn': 'no column specified',
		'setquery.error.notable': 'no table specified',
		'savecsv.error.fileexists': 'File already exists',
		'savecsv.confirm.schemaini': "schema.ini に {0} に関する指定が残っています.\nこのまま保存するとこの指定によって出力が変化します.\nそれでも保存を行いますか ?"
	};
	if (!$.DataSource) {
		$.DataSource = {};
	}
	var CSVDB = $.DataSource.CSVDB = function(path) {
		if (this instanceof CSVDB) {
			this.open(path);
			return this;
		}
		return new CSVDB(path);
	};

	CSVDB.prototype.open = function(path) {
		var fso = new ActiveXObject("Scripting.FileSystemObject");
		if (!fso.FileExists(path)) {
			fso = null;
			throw new Error(messages['open.error.filenotexist']);
		}
		fso = null;
		this._path = path;
		var sep = path.lastIndexOf('\\');
		this._dir = sep < 0 ? new ActiveXObject('WScript.Shell').CurrentDirectory.toString() : path.substring(0, sep);
		this._filename = sep < 0 ? path : path.substring(sep + 1);
		this._pageSize = 10;
		this._currentPage = 1;
		this._queryStruct = {
			statement: "SELECT",
			predicate: "",
			columns: [ "*" ],
			tables: [ this._filename ],
			where: "",
			groupby: "",
			having: "",
			orderby: [],
			withoOwnerAccess: false
		};
		this._query = this._buildQueryString();
		var conn = null;
		try {
			this._connection = conn = new ActiveXObject('ADODB.Connection');
			conn.Open('Driver={Microsoft Text Driver (*.txt; *.csv)};DBQ=' + this._dir + ';');
			if (conn.State == 0)
				throw new Error(messages['open.error.connection']);
		} catch (e) {
			if (conn != null) {
				conn.Close();
				this._connection = conn = null;
			}
			throw e;
		}
	};

	CSVDB.prototype.close = function() {
		if (this._connection) {
			this._connection.Close();
			this._connection = null;
		}
	};

	CSVDB.prototype.setPageSize = function(size) {
		this._pageSize = size >= 1 ? size : 1;
	};

	CSVDB.prototype.setCurrentPage = function(cur) {
		this._currentPage = cur >= 1 ? cur : 1;
	};

	CSVDB.prototype.setSort = function(sortparams) {
		var orderby = [];
		for (var i = 0; i < sortparams.length; i++) {
			var c = sortparams[i].column;
			if (c) {
				var o = sortparams[i].order;
				orderby.push({ column: c, order: (o && o.toUpperCase && o.toUpperCase() == "DESC") ? "DESC" : "ASC" });
			}
		}
		this._queryStruct.orderby = orderby;
		this._query = this._buildQueryString();
	};

	CSVDB.prototype.setQuery = function(sql) {
		this._queryStruct = this._parseQuery(sql);
		this._query = this._buildQueryString();
	};

	CSVDB.prototype.getQuery = function() {
		return new String(this._query);
	};

	CSVDB.prototype._maskNames = function(sql) {
		var c, inname = false, start = 0, end = 0, parts = [];
		for (var i = 0, len = sql.length; i < len; i++) {
			c = sql.charAt(i);
			if (c == '[') {
				if (!inname) {
					parts.push( sql.substring(end, i + 1) );
					start = i + 1;
					inname = true;
				}
			} else if (c == ']') {
				if (inname) {
					parts.push( $.map(new Array(i - start), function(){ return '_'; }).join('') );
					end = i;
					inname = false;
				} else {
					throw new Error(messages['setquery.error.brackets']);
				}
			}
		}
		if (inname) {
			throw new Error(messages['setquery.error.brackets']);
		}
		parts.push(sql.substring(end));
		return parts.join('');
	}

	CSVDB.prototype._maskSubQuery = function(sql) {
		sql = this._maskNames(sql);
		var c, depth = 0, start = 0, end = 0, parts = [];
		for (var i = 0, len = sql.length; i < len; i++) {
			c = sql.charAt(i);
			if (c == '(') {
				if (++depth == 1) {
					parts.push( sql.substring(end, i + 1) );
					start = i + 1;
				}
			} else if (c == ')') {
				if (--depth == 0) {
					parts.push( $.map(new Array(i - start), function(){ return '_'; }).join('') );
					end = i;
				} else if (depth < 0) {
					throw new Error(messages['setquery.error.parentheses']);
				}
			}
		}
		if (depth != 0) {
			throw new Error(messages['setquery.error.parentheses']);
		}
		parts.push(sql.substring(end));
		return parts.join('');
	}

	CSVDB.prototype._parseQuery = function(sql) {
		var ret = {}, m = null, sqlmask;

		sql = $.trim(sql);
		if (sql.length == 0) {
			throw new Error(messages['setquery.error.emptystmt']);
		}

		if (!/^SELECT\s/i.test(sql)) {
			throw new Error(messages['setquery.error.notselect']);
		}
		ret.statement = "SELECT";
		sql = $.trim(sql.substring(7));

		if (m = /^(ALL|DISTINCT|DISTINCTROW|TOP)\b/i.exec(sql)) {
			ret.predicate = m[1].toUpperCase();
			sql = $.trim(sql.substring(m[0].length));
		} else {
			ret.predicate = "";
		}

		if ((m = /\bFROM\b/i.exec(this._maskSubQuery(sql))) == null) {
			throw new Error(messages['setquery.error.nofrom']);
		}
		ret.columns = $.trim(sql.substring(0, m.index));
		if (ret.columns.length == 0) {
			throw new Error(messages['setquery.error.nocolumn']);
		}
		ret.columns = ret.columns.split(/\s*,\s*/);
		sql = $.trim(sql.substring(m.index + m[0].length));

		if (m = /\b(WHERE|GROUP|HAVING|ORDER|WITH)\b/i.exec(this._maskSubQuery(sql))) {
			ret.tables = $.trim(sql.substring(0, m.index));
			sql = $.trim(sql.substring(m.index));
		} else {
			ret.tables = $.trim(sql);
			sql = "";
		}
		if (ret.tables.length == 0) {
			throw new Error(messages['setquery.error.notable']);
		}
		ret.tables = ret.tables.split(/\s*,\s*/);

		if (m = /^WHERE\b/i.exec(sql)) {
			sql = $.trim(sql.substring(5));
			if (m = /\b(GROUP|HAVING|ORDER|WITH)\b/i.exec(this._maskSubQuery(sql))) {
				ret.where = $.trim(sql.substring(0, m.index));
				sql = sql.substring(m.index);
			} else {
				ret.where = $.trim(sql);
				sql = "";
			}
		} else {
			ret.where = "";
		}

		if (m = /^GROUP\s+BY\b/i.exec(sql)) {
			sql = $.trim(sql.substring(m[0].length));
			if (m = /\b(HAVING|ORDER|WITH)\b/i.exec(this._maskSubQuery(sql))) {
				ret.groupby = $.trim(sql.substring(0, m.index));
				sql = sql.substring(m.index);
			} else {
				ret.groupby = $.trim(sql);
				sql = "";
			}
		} else {
			ret.groupby = "";
		}

		if (m = /^HAVING\b/i.exec(sql)) {
			sql = $.trim(sql.substring(6));
			if (m = /\b(ORDER|WITH)\b/i.exec(this._maskSubQuery(sql))) {
				ret.having = $.trim(sql.substring(0, m.index));
				sql = sql.substring(m.index);
			} else {
				ret.having = $.trim(sql);
				sql = "";
			}
		} else {
			ret.having = "";
		}

		if (m = /^ORDER\s+BY\b/i.exec(sql)) {
			sql = $.trim(sql.substring(m[0].length));
			if (m = /\bWITH\b/i.exec(this._maskSubQuery(sql))) {
				ret.orderby = $.trim(sql.substring(0, m.index));
				sql = sql.substring(m.index);
			} else {
				ret.orderby = $.trim(sql);
				sql = "";
			}
			ret.orderby = $.map(ret.orderby.split(/\s*,\s*/), function(s, i){
				var m = null;
				if (m = /\b(ASC|DESC)\s*$/i.exec()) {
					if (m[1].toUpperCase() == "DESC") {
						return { column: $.trim(s.substring(0, m.index)), order: "DESC" };
					} else {
						return { column: $.trim(s.substring(0, m.index)), order: "ASC" };
					}
				} else {
					return { column: $.trim(s), order: "ASC" };
				}
			});
		} else {
			ret.orderby = [];
		}

		if (/^WITH\s+OWNERACCESS\s+OPTION\b/i.test(sql)) {
			ret.withOwnerAccess = true;
		} else {
			ret.withOwnerAccess = false;
		}

		return ret;
	}

	CSVDB.prototype._buildQueryString = function(into) {
		// "{" が列名に含まれているとエラー
		var s = this._queryStruct;
		var q = s.statement;
		if (s.predicate.length) {
			q += " " + s.predicate;
		}
		q += " " + s.columns.join(", ");
		if (into) {
			q += " INTO " + into;
		}
		q += "\n FROM " + s.tables.join(", ");
		if (s.where.length) {
			q += "\n WHERE " + s.where;
		}
		if (s.groupby.length) {
			q += "\n GROUP BY " + s.groupby;
		}
		if (s.having.length) {
			q += "\n HAVING " + s.having;
		}
		if (s.orderby.length) {
			q += "\n ORDER BY " + $.map(s.orderby, function(s, i){ return "[" + s.column + "] " + s.order; }).join(", ");
		}
		if (s.withoOwnerAccess) {
			q += "\n WITH OWNERACCESS OPTION";
		}
		return q;
	};

	CSVDB.prototype.header = function() {
		var recordset = null;
		try {
			recordset = new ActiveXObject("ADODB.RecordSet");
			recordset.Open(this._query, this._connection);
			var ret = [];
			for (var e = new Enumerator(recordset.Fields); !e.atEnd(); e.moveNext()) {
				var item = e.item();
				ret.push({ name: '' + item.Name, type: item.Type, precision: item.Precision, numericscale: item.NumericScale });
			}
			return ret;
		} finally {
			if (recordset != null) {
				recordset.Close();
				recordset = null;
			}
		}
	};

	CSVDB.prototype.json = function() {
		var recordset = null;
		try {
			recordset = new ActiveXObject("ADODB.RecordSet");
			recordset.Open(this._query, this._connection, 3, 1, 1); // adOpenKeyset, adLockReadOnly, adCmdText
			var size = recordset.PageSize = this._pageSize;
			var page = recordset.AbsolutePage = this._currentPage;
			var total = recordset.PageCount;
			var start = (page - 1) * size + 1;
			var rows = [];
			for (var count = 0; !recordset.EOF && count < size; recordset.MoveNext(), count++) {
				var fields = [];
				for (var e = new Enumerator(recordset.Fields); !e.atEnd(); e.moveNext()) {
					fields.push('' + e.item().Value); // ここで文字列に変換しておかないと null が空白になる
				}
				rows.push( '{"id":"' + (start + count) + '","cell":[' + (fields.length ? '"' + fields.join('","') + '"' : '') + ']}' );
			}
			return '{' +
				'"total":' + total + ',' +
				'"page":' + this._currentPage + ',' +
				'"records":' + recordset.RecordCount + ',' +
				'"rows":[' + rows.join(',') + ']' +
				'}';
		} finally {
			if (recordset != null) {
				recordset.Close();
				recordset = null;
			}
		}
	};

	CSVDB.prototype.saveCsv = function(path, window) {
		var fso = new ActiveXObject("Scripting.FileSystemObject");
		if (/^([a-zA-Z]:)|(\\\\)/.test(path)) {
			// absolute
		} else {
			path = fso.GetAbsolutePathName(this._dir + "\\" + path);
		}

		if (fso.FileExists(path)) {
			throw new Error(messages['savecsv.error.fileexists']);
		}

		var sep = path.lastIndexOf('\\');
		var dir = path.substring(0, sep);
		var file = path.substring(sep + 1);

		if (fso.FileExists(dir + "\\schema.ini")) {
			try {
				var regex = new RegExp("^\\[" + file + "\\]");
				var found = false;
				$.each(fso.OpenTextFile(dir + "\\schema.ini").ReadAll().split("\n"), function(){
					if (regex.test(this)) {
						found = true;
						return false;
					}
				});
				if (found) {
					if (window) {
						if (!window.confirm(messages['savecsv.confirm.schemaini'].replace("{0}", file)))
							return;
					} else {
						return;
					}
				}
			} catch (e) {
			}
		}

		if (dir == this._dir) {
			var query = this._buildQueryString('[' + file + ']');
		} else {
			var query = this._buildQueryString('[' + file + '] IN "' + dir + '" "Text;"'); // may be "Text;HDR=YES;"
		}
		this._connection.Execute(query, null, 0x81); // adCmdText | adExecuteNoRecords
	};

	CSVDB.prototype.getSourceName = function() {
		return this._queryStruct.tables.join(', ');
	};

	CSVDB.prototype.supports = function(method) {
		switch (method) {
		case 'setQuery':
		case 'getQuery':
		case 'setSort':
		case 'saveCsv':
		case 'getSourceName':
			return true;
		}
		return false;
	};

})(jQuery);
