/**
 * @file     dscsvfs.js
 * @description
 *           CSVファイルからJSONで読み出す.
 *           CSVはRFC4180の拡張版を扱う.
 * @auther   Masashi Nakanishi
 * @license  MIT License  http://www.opensource.org/licenses/mit-license.php
 */
;(function($){
	if (!$.DataSource) {
		$.DataSource = {};
	}
	var CSVFile = $.DataSource.CSVFile = function(path) {
		if (this instanceof CSVFile) {
			this.open(path);
			return this;
		}
		return new CSVFile(path);
	};

	CSVFile.prototype.open = function(path, headers) {
		var fso = new ActiveXObject("Scripting.FileSystemObject");
		var stream = null;
		try {
			stream = fso.OpenTextFile(file, 1, true);
			var lines = [];
			while (!stream.AtEndOfStream) {
				lines.push(stream.ReadLine());
			}
			// remove the last empty line if exists
			if (lines.length && lines[lines.length - 1] == '') {
				lines.splice(lines.length - 1, 1);
			}
			// merge lines because of newlines in cells
			for (var i = 0; i < lines.length; i++) {
				var m = lines[i].match(/"/g);
				if (m != null && m.length % 2 != 0) {
					if (i == lines.length - 1)
						throw new Error('parse error');
					lines[i] += "\n" + lines[i + 1];
					lines.splice(i + 1, 1);
					i--;
				}
			}
			// split a line to cells
			for (var i = 0; i < lines.length; i++) {
				var cells = lines[i].split(",");
				for (var k = 0; k < cells.length; k++) {
					var cell = cells[k];
					var m = cell.match(/"/g);
					if (m != null && m.length % 2 != 0) {
						cells[k] += "," + cells[k + 1];
						lines.splice(k + 1, 1);
						k--;
						continue;
					}
					if (cell.length >= 2) {
						if (cell.charAt(0) == '"' && cell.charAt(cell.length - 1) == '"') {
							cells[k] = cell.substring(1, cell.length - 1).replace(/""/g, '"'); // unescape
						}
					}
				}
				lines[i] = cells;
			}
			this._rows = lines;
		} finally {
			if (stream) {
				stream.Close();
				stream = null;
			}
			fso = null;
		}

		this._headerPresent = true; // default is true
		if (headers instanceof Array) {
			this._headerPresent = false;
			this._headerArray = headers;
		}
		this._path = path;
		this._pageSize = 10;
		this._currentPage = 1;
		this._sortColumn = null;
		this._sortOrder = "ASC";
	};

	CSVFile.prototype.close = function() {
		this._rows = null;
		this._headerArray = null;
	};

	CSVFile.prototype.setPageSize = function(size) {
		this._pageSize = size >= 1 ? size : 1;
	};

	CSVFile.prototype.setCurrentPage = function(cur) {
		this._currentPage = cur >= 1 ? cur : 1;
	};

	CSVFile.prototype.setSort = function(sortparams) {
		if (sortparams.length) {
			this._sortColumn = sortparams[0].column;
			this._sortOrder = (sortparams[0].order && sortparams[0].order.toUpperCase() == "DESC") ? "DESC" : "ASC";
		} else {
			this._sortColumn = null;
			this._sortOrder = "ASC";
		}
	};

	CSVFile.prototype.setQuery = function(sql) {
		//TODO
	};
	CSVFile.prototype.getQuery = function(sql) {
		//TODO
	};

	CSVFile.prototype.header = function() {
		if (this._headerPresent) {
			var cells = rows[0];
			var ret = [];
			for (var i = 0, n = cells.length; i < n; i++) {
				ret[i] = { name: cells[i] };
			}
			return ret;
		} else if (this._headerArray) {
			var arr = this._headerArray;
			var ret = [];
			for (var i = 0, n = arr.length; i < n; i++) {
				ret[i] = { name: arr[i] };
			}
			return ret;
		} else {
			if (rows.length == 0)
				return [];
			var ncol = rows[0].length;
			var ret = [];
			for (var i = 1; i <= ncol; i++) {
				ret[i - 1] = { name: '' + i };
			}
			return ret;
		}
	};

	CSVFile.prototype.json = function() {
		//TODO sort
		var pageSize = this._pageSize;
		var currentPage = this._currentPage;
		var start = (currentPage - 1) * pageSize;
		var end = currentPage * pageSize;
		if (this._headerPresent) {
			start += 1;
			end += 1;
		}
		var rows = this._rows;
		var json = "";
		for (var i = 0, nrows = rows.length; i < nrows; i++) {
			if (i < start) {
				continue; // skip
			}
			var cells = rows[i];
			json += (json.length ? ',' : '') + '{"id":"' + (i + 1) + '","cell":[' + (cells.length ? '"' + cells.join('","') + '"' : '') + ']}';
			if (end > start && i + 1 >= end) {
				break;
			}
		}
		var total = Math.ceil(rows.length / pageSize);
		return '{' +
			'"total":' + total + ',' +
			'"page":' + currentPage + ',' +
			'"records":' + rows.length + ',' +
			'"rows":[' + json + ']' +
			'}';
	};

	CSVFile.prototype.saveCsv = function(path, window) {
		//TODO
	};

	CSVFile.prototype.supports = function(method) {
		switch (method) {
		//case 'setQuery':
		//case 'getQuery':
		//case 'setSort':
		//case 'saveCsv':
		//case 'getSourceName':
		//	return true;
		}
		return false;
	};

})(jQuery);
